import abc


class BaseAgent(abc.ABC):
    @abc.abstractmethod
    def generate_audio(self, prompt, **kwargs):
        pass

    @abc.abstractmethod
    def generate_image(self, prompt, **kwargs):
        pass

    @abc.abstractmethod
    def generate_json(
        self, text, functions, name="json_agent", primer_text="", file=None, context=[]
    ):
        pass

    @abc.abstractmethod
    def generate_text(
        self, text, name="text_agent", primer_text="", file=None, context=[]
    ):
        pass

    @abc.abstractmethod
    def summarize_text(self, text, primer_text=""):
        pass



import json
import os

import requests
from slugify import slugify

from autonomous import log

from .page import WikiJSPage


class WikiJS:
    endpoint = f"{os.environ.get('WIKIJS_BASE_URL')}/{os.environ.get('WIKIJS_API_PATH', 'graphql')}"
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {os.environ.get('WIKIJS_TOKEN')}",
    }

    @classmethod
    def make_request(cls, query, obj_vars=None):
        # log(query)
        # log(**obj_vars)

        variables = json.dumps(obj_vars)

        response = requests.post(
            WikiJS.endpoint,
            headers=cls.headers,
            json={"query": query, "variables": variables},
        )
        if response.status_code != 200:
            raise Exception(response.text)

        # log(response.text)

        return response

    @classmethod
    def get_page(cls, id):
        id = int(id)
        query = """
        query($id:Int!){
            pages{
                single(id:$id){
                    id
                    path
                    title
                    content
                    updatedAt
                    tags {
                        title
                    }
                    description
                }
            }
        }
        """
        res = cls.make_request(query, {"id": id})
        # log(res, res.text)
        return (
            WikiJSPage(**(res.json()["data"]["pages"]["single"]))
            if res.json()["data"]["pages"]["single"]
            else None
        )

    @classmethod
    def pull_tagged(cls, tags=None):
        query = """
            query($tags:[String!]!){
                pages{
                    list (tags: $tags){
                        id
                        path
                        title
                    }
                }
            }
        """
        # log(query)
        response = cls.make_request(query, {"tags": tags})
        # log(response.text)
        results = response.json()["data"]["pages"]["list"]
        # log(results)
        pages = [cls.get_page(p["id"]) for p in results]
        return pages

    @classmethod
    def create_page(cls, title, content, description, tags, path):
        if page := cls.find_by_path(path):
            return cls.update_page(
                page.id,
                title=title,
                content=content,
                description=description,
                tags=tags,
            )

        obj_vars = {
            "title": title,
            "content": content,
            "description": description,
            "tags": tags,
            "path": path,
        }
        query = """
        mutation($description:String!, $title:String!, $content:String!, $path:String!, $tags:[String!]!){
            pages
            {
                create (
                    content: $content,
                    description: $description,
                    title: $title,
                    tags: $tags,
                    path: $path,
                    isPublished: true,
                    isPrivate: false,
                    editor: "markdown",
                    locale: "en"
                    )
                {
                    page{
                        id
                        path
                        title
                    }
                    responseResult{
                        message
                        succeeded
                        errorCode
                    }
                }
            }
        }
        """

        response = cls.make_request(query, obj_vars)
        results = response.json()
        page = results["data"]["pages"]["create"]["page"]
        if not page:
            # log(f"results: {response.json()['data']['pages']['create']['responseResult']}")
            return response.json()["data"]["pages"]["create"]["responseResult"]
        result = WikiJSPage(**page)
        return result

    @classmethod
    def update_page(cls, id, **kwargs):
        """
        params:
        - name: the name of the object
        - desc: the description of the object
        - wikijs_id: the id of the object in wikijs (**must provide either id or path**)
        - path: the path to the object in wikijs (**must provide either id or path**)
        - tags: a list of tags to add to the object
        """
        page = cls.get_page(id)
        obj_vars = {
            "id": id,
            "title": kwargs.get("title", page.title),
            "description": kwargs.get("description", page.description),
            "content": kwargs.get("content", page.content),
            "tags": kwargs.get("tags", page.tags),
        }
        query = """
        mutation($id:Int!, $description:String!, $title:String!, $content:String!, $tags:[String!]!){
            pages
            {
                update (
                    id: $id,
                    content: $content,
                    description: $description,
                    title: $title,
                    tags: $tags,
                    )
                {
                    page{
                        id
                        path
                        title
                    }
                    responseResult{
                        message
                        succeeded
                        errorCode
                    }
                }
            }
        }
        """

        res = cls.make_request(query, obj_vars)
        results = res.json()["data"]["pages"]["update"]["page"]
        if not results:
            # log(f"results: {res.json()['data']['pages']['create']['responseResult']}")
            return
        result = WikiJSPage(**results)
        return result

    @classmethod
    def find_by_title(cls, title):
        query = """
            query($title:String!){
                pages
                {
                    search (query: $title)
                    {
                        results {
                            id
                            path
                            title
                        }
                    }
                }
            }
            """
        # log(query)
        response = cls.make_request(query, {"title": title})
        # log(response.text)
        results = response.json()["data"]["pages"]["search"]["results"]

        try:
            for p in results:
                # log(title, p)
                if title == p["title"]:
                    return int(p["id"])
        except KeyError as e:
            log(e)
            log(response.text)
        return None

    @classmethod
    def find_by_path(cls, path):
        if path.startswith("/"):
            path = path[1:]
        query = """
            query($path:String!){
                pages
                {
                    singleByPath (path: $path, locale: "en")
                    {
                        id
                        path
                        title
                    }
                }
            }
            """
        response = cls.make_request(query, {"path": path})
        # log(response.json())
        results = response.json()["data"]["pages"]["singleByPath"]
        return WikiJSPage(**results) if results else None

    @classmethod
    def remove_page(cls, id):
        # log(query)

        query = """
        mutation($id:Int!){
            pages{
                delete(id:$id){
                    responseResult{
                        message
                        errorCode
                        succeeded
                    }
                }
            }
        }
        """
        res = cls.make_request(query, {"id": id})
        return res.json()["data"]["pages"]["delete"]["responseResult"]["succeeded"]



import os

import github
from dotenv import dotenv_values, load_dotenv

from .GHRepo import Repository


class GHOrganization:
    def __init__(self, org=None, token=None, username=None, email=None, name=None):
        """
        _summary_

        Args:
            org (_type_, optional): _description_. Defaults to None.
        """
        self.token = token or os.getenv("GITHUB_PAT")
        self.org = org or os.getenv("GITHUB_ORG")
        self.username = username or os.getenv("GITHUB_USERNAME")
        self.email = email or os.getenv("GITHUB_EMAIL")
        self.commit_name = name or os.getenv("GITHUB_COMMIT_NAME")
        # print(vars(self), os.getenv('GITHUB_PAT'))
        if self.token and self.org:
            self.g = github.Github(self.token)
            self.org = self.g.get_organization(self.org) if self.org else None
        self.repositories = []

    def get_repos(self, destination="./", search=""):
        """
        _summary_
        """
        for repo in self.org.get_repos():
            if search in repo.name:
                self.repositories.append(Repository(repo, destination))
        return self.repositories



class AutoAttribute:
    def __init__(
        self, type, default=None, required=False, unique=False, primary_key=False
    ):
        if type in [
            "TEXT",
            "NUMERIC",
            "MODEL",
        ]:
            self.type = type
        else:
            raise ValueError(f"Invalid type {type}")

        self.default = default
        self.required = required
        self.unique = unique
        self.primary_key = primary_key

    def __repr__(self):
        return f"<AutoAttribute {self.type} {self.default} {self.required} {self.unique} {self.primary_key}>"



import inspect
import logging
import os

# class LogFilter(logging.Filter):
#     def filter(self, record):
#         return record.levelno == LOG_LEVEL
log_levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


class Logger:
    """
    Set the log level to by setting the LOG_LEVEL environment variable to:
    "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"

    Default: WARNING
    """

    def __init__(self):
        self.logger = logging.getLogger("gunicorn.error")
        level = os.environ.get("LOG_LEVEL") or self.logger.level
        self.logger.setLevel(log_levels.get(level, "DEBUG"))
        self.enabled = True

    def set_level(self, level):
        self.logger.setLevel(log_levels[level])

    def enable(self, enable=True):
        self.enabled = enable

    def __call__(self, *args, **kwargs):
        if self.enabled:
            caller = inspect.stack()[1]
            fn = caller.filename.split("/")[-1]
            msg = f"\n\n{'='*20}\t{fn}:{caller.function}()::{caller.lineno}\t{'='*20}\n"
            msg += f"\n{'='*20}\n"
            if args:
                msg += "\n---\n".join([f"{v}" for v in args])
            if kwargs:
                msg += "\n---\n".join([f"{k} : {v}" for k, v in kwargs.items()])
            msg += f"\n{'='*20}\n"
            self.logger.log(self.logger.level, f"{msg}\n")


log = Logger()



from datetime import datetime

from autonomous import log
from autonomous.db.autodb import Database


class ORM:
    _database = Database()

    def __init__(self, name, attributes):
        self.table = self._database.get_table(table=name, schema=attributes)
        self.name = name

    def _replace_pk_with_id(self, data):
        if isinstance(data, dict):
            for key in list(data.keys()):
                if key == "pk":
                    data["_id"] = data.pop("pk")
                else:
                    self._replace_pk_with_id(data[key])
                    # breakpoint()
        elif isinstance(data, list):
            for item in data:
                self._replace_pk_with_id(item)

    def _replace_id_with_pk(self, data):
        if isinstance(data, dict):
            for key in list(data.keys()):
                if key == "_id":
                    data["pk"] = data.pop("_id")
                else:
                    self._replace_id_with_pk(data[key])
                    # breakpoint()
        elif isinstance(data, list):
            for item in data:
                self._replace_id_with_pk(item)

    def save(self, data):
        self._replace_pk_with_id(data)
        if result := self.table.save(data):
            self._replace_id_with_pk(result)
        return result

    def get(self, pk):
        pk = str(pk)
        if result := self.table.get(pk):
            self._replace_id_with_pk(result)
        return result

    def all(self):
        if results := self.table.all():
            self._replace_id_with_pk(results)
        return results

    def search(self, **kwargs):
        self._replace_pk_with_id(kwargs)
        if results := self.table.search(**kwargs):
            self._replace_id_with_pk(results)
        return results

    def find(self, **kwargs):
        log(kwargs)
        self._replace_pk_with_id(kwargs)
        log(kwargs)
        if result := self.table.find(**kwargs):
            self._replace_id_with_pk(result)
        log(result)
        return result

    def random(self):
        if result := self.table.random():
            self._replace_id_with_pk(result)
        return result

    def delete(self, pk):
        pk = str(pk)
        return self.table.delete(_id=pk)

    def flush_table(self):
        return self.table.clear()

    def dbdump(self, directory="./dbbackups"):
        return self._database.dbdump(directory)

    def dbload(self, directory="./dbbackups"):
        return self._database.dbload(directory)



import json
import os
from base64 import b64decode

from openai import OpenAI

from autonomous import log


class OpenAIAgent:
    client = None

    _instructions_addition = """
        IMPORTANT: always use the function 'response' tool to respond to the user with the requested JSON schema. Never add any other text to the response.
        """

    def __init__(self, **kwargs):
        self.client = OpenAI(api_key=os.environ.get("OPENAI_KEY"))

    def generate_audio(self, prompt, **kwargs):
        image = None
        try:
            response = self.client.images.generate(
                model="dall-e-3", prompt=prompt, response_format="b64_json", **kwargs
            )
            image_dict = response.data[0]
        except Exception as e:
            log(f"==== Error: Unable to create image ====\n\n{e}")
        else:
            image = b64decode(image_dict.b64_json)
        return image

    def generate_image(self, prompt, **kwargs):
        image = None
        try:
            response = self.client.images.generate(
                model="dall-e-3", prompt=prompt, response_format="b64_json", **kwargs
            )
            image_dict = response.data[0]
        except Exception as e:
            log(f"==== Error: Unable to create image ====\n\n{e}")
        else:
            image = b64decode(image_dict.b64_json)
        return image

    def generate_json(
        self, text, function, name=None, primer_text="", file_data=None, context=[]
    ):
        json_data = {
            "response_format": {"type": "json_object"},
            "messages": [
                {
                    "role": "system",
                    "content": f"{primer_text}. {self._instructions_addition}",
                },
                {
                    "role": "user",
                    "content": text,
                },
            ],
            "tools": [{"type": "function", "function": function}],
            "tool_choice": {"type": "function", "function": {"name": function["name"]}},
        }

        try:
            response = self.client.chat.completions.create(
                model="gpt-4-turbo-preview", **json_data
            )
        except Exception as e:
            log(f"{type(e)}:{e}\n\n==== Error: fall back to lesser model ====")
            response = self.client.chat.completions.create(model="gpt-4", **json_data)
        # breakpoint()
        try:
            results = response.message.tool_calls[0].function.arguments
            results = results[results.find("{") : results.rfind("}") + 1]
            log(results)
            json_result = json.loads(results)
        except Exception as e:
            log(f"==== Unable to generate content ====\n\n{type(e)}:{e}")
            return None

        return json_result

    def generate_text(
        self, text, name="text_agent", primer_text="", file_data=None, context=[]
    ):
        json_data = {
            "messages": [
                {
                    "role": "system",
                    "content": primer_text,
                },
                {
                    "role": "user",
                    "content": text,
                },
            ]
        }

        try:
            response = self.client.chat.completions.create(model="gpt-4", **json_data)
        except Exception as e:
            log(f"{type(e)}:{e}\n\n==== Error: fall back to lesser model ====")
            response = self.client.chat.completions.create(
                model="gpt-3.5-turbo", **json_data
            )
        # breakpoint()
        try:
            result = response.choices[0].message.content
        except Exception as e:
            log(f"==== Unable to generate content ====\n\n{type(e)}:{e}")
            return None

        return result

    def summarize_text(self, text, primer=""):
        message = [
            {
                "role": "system",
                "content": f"You are a highly skilled AI trained in language comprehension and summarization.{primer}",
            },
            {"role": "user", "content": text},
        ]
        try:
            response = self.client.chat.completions.create(
                model="gpt-4", temperature=0, messages=message
            )
        except Exception as e:
            log(f"{type(e)}:{e}\n\n==== Error: fall back to lesser model ====")
            response = self.client.chat.completions.create(
                model="gpt-4", temperature=1, messages=message
            )
        # breakpoint()

        try:
            result = response.choices[0].message.content
        except Exception as e:
            log(f"{type(e)}:{e}\n\n Unable to generate content ====")
            return e

        return result



import os
from base64 import b64decode

from autonomous import log
from langchain.chat_models import ChatOpenAI
from langchain.prompts import ChatPromptTemplate, PromptTemplate
from langchain.schema import HumanMessage, SystemMessage
from openai import OpenAI


class LangChainAgent:
    client = None
    # msg_types = {"human": HumanMessage, "system": "system", "agent": "agent"}

    def __init__(self, **kwargs):
        self.client = ChatOpenAI(api_key=os.environ.get("OPENAI_KEY"))
        self.messages = []
        self.agents = []

    def add_message(self, content, msg_type="human"):
        msg = self.msg_types.get(msg_type, HumanMessage)(content=content)
        self.messages.append(msg)
        return msg

    def add_agent(self, agent):
        self.client = agent

    # ************************************************************#
    # *  Generation Methods                                      *#
    # ************************************************************#

    def generate_image(self, prompt="", **kwargs):
        if prompt:
            self.add_message(prompt)

        image = None
        try:
            response = self.client.images.generate(
                model="dall-e-3", prompt=prompt, response_format="b64_json", **kwargs
            )
            image_dict = response.data[0]
        except Exception as e:
            log(f"==== Error: Unable to create image ====\n\n{e}")
        else:
            image = b64decode(image_dict.b64_json)
        return image

    def generate_json(self, text, functions, primer_text=""):
        json_data = {
            # "response_format":{ "type": "json_object" },
            "messages": [
                {
                    "role": "system",
                    "content": f"{primer_text}. Your output must be a JSON object.",
                },
                {
                    "role": "user",
                    "content": text,
                },
            ]
        }

        if isinstance(functions, (list, tuple)):
            json_data.update({"functions": functions})
        elif functions is not None:
            json_data.update({"function_call": {"name": functions["name"]}})
            json_data.update({"functions": [functions]})

        # try:
        response = self.client.chat.completions.create(model="gpt-4", **json_data)
        # except Exception as e:
        #     log(f"{type(e)}:{e}\n\n==== Error: fall back to lesser model ====")
        #     response = self.client.chat.completions.create(
        #         model="gpt-3.5-turbo", **json_data
        #     )
        # breakpoint()
        try:
            result = response.choices[0].message.function_call.arguments
        except Exception as e:
            log(f"==== Unable to generate content ====\n\n{type(e)}:{e}")
            return None

        return result

    def generate_text(self, prompt="", primer_text=""):
        if prompt:
            self.add_message(prompt)
        json_data = {
            "messages": [
                {
                    "role": "system",
                    "content": primer_text,
                },
                {
                    "role": "user",
                    "content": prompt,
                },
            ]
        }

        try:
            response = self.client.predict_messages(self.messages)
            self.messages = []
            result = response.content
        except Exception as e:
            log(f"==== Unable to generate content ====\n\n{type(e)}:{e}")
            return None
        else:
            self.messages = []
        return result

    def summarize_text(self, text):
        prompt = [
            SystemMessage(
                content="You are a highly skilled AI trained in language comprehension and summarization. You will summarize text ensuring that you maintain the key elements of the otiginal and as well as any important context.",
                additional_kwargs={},
            ),
            HumanMessage(
                content=f"Summarize the following text: {text}", example=False
            ),
        ]

        try:
            response = self.client.predict_messages(prompt)
            result = response.content
        except Exception as e:
            log(f"{type(e)}:{e}\n\n Unable to generate content ====")
            return e
        else:
            self.messages = []

        return result



import io
import os
import shutil
import uuid

from PIL import Image

from autonomous import log


class ImageStorage:
    _sizes = {"thumbnail": 100, "small": 300, "medium": 600, "large": 1000}

    def __init__(self, path="static/images"):
        self.base_path = path

    @classmethod
    def _get_key(cls, folder="", pkey=None):
        if folder and not folder.endswith("/"):
            folder = f"{folder}/"
        folder = f"{folder.replace('/', '.')}{pkey or uuid.uuid4()}"
        return f"{folder}"

    def _resize_image(self, asset_id, max_size=1024):
        # log("Resizing image", asset_id, max_size)
        file_path = f"{self.get_path(asset_id)}/orig.webp"
        with Image.open(file_path) as img:
            resized_img = img.copy()
            max_size = self._sizes.get(max_size) or int(max_size)
            resized_img.thumbnail((max_size, max_size))
            img_byte_arr = io.BytesIO()
            resized_img.save(img_byte_arr, format="WEBP")
            return img_byte_arr.getvalue()

    def _convert_image(self, raw, crop=False):
        with Image.open(io.BytesIO(raw)) as img:
            width, height = img.size
            width, height = img.size
            if crop and width != height:
                max_size = min(width, height)
                img.crop(
                    (
                        (width - max_size) / 2,
                        (height - max_size) / 2,
                        (width + max_size) / 2,
                        (height + max_size) / 2,
                    )
                )
            img = img.copy()
            img_byte_arr = io.BytesIO()
            img.save(img_byte_arr, format="WEBP")
            return img_byte_arr.getvalue()

    def save(self, file, folder="", crop=False):
        asset_id = self._get_key(folder)
        os.makedirs(self.get_path(asset_id), exist_ok=True)
        file_path = f"{self.get_path(asset_id)}/orig.webp"
        with open(file_path, "wb") as asset:
            file = self._convert_image(file, crop=crop)
            asset.write(file)
        return asset_id

    def get_url(self, asset_id, size="orig", full_url=False):
        # log(f"Getting image: {asset_id}.{size}")
        if not asset_id:
            return ""
        original_path = f"{self.get_path(asset_id)}"
        # log(f"Getting image: {asset_id}.{size}", original_path)
        if not os.path.exists(original_path):
            # log(f"Original image not found: {original_path}")
            return ""
        file_path = f"{original_path}/{size}.webp"
        # log(file_path)
        if (
            size != "orig"
            and os.path.exists(original_path)
            and not os.path.exists(file_path)
        ):
            # If the file doesn't exist, create it
            result = self._resize_image(asset_id, size)
            with open(file_path, "wb") as asset:
                asset.write(result)
        result_url = (
            f"/{file_path}"
            if not full_url
            else f"{os.environ.get('APP_BASE_URL', '')}/{file_path}"
        )
        # log(f"Returning image url: {result_url}")
        return result_url

    def get_path(self, asset_id):
        if asset_id:
            asset_path = asset_id.replace(".", "/")
            if asset_path.endswith("/"):
                asset_path = asset_path[:-1]
            return os.path.join(self.base_path, f"{asset_path}")
        else:
            return self.base_path

    def search(self, folder="", **kwargs):
        imgs = []
        # log(f"{self.base_path}")
        for f in os.listdir(f"{self.base_path}/{folder}"):
            # log(f"{self.base_path}/{folder}/{f}")
            for img in os.listdir(f"{self.base_path}/{folder}/{f}"):
                img_key = self._get_key(
                    f"{folder}",
                    pkey=f,
                )
                imgs.append(img_key)
        # log(imgs)
        return imgs

    def remove(self, asset_id):
        file_path = self.get_path(asset_id)
        if os.path.isdir(file_path):
            shutil.rmtree(file_path)
            return True
        return False



from datetime import datetime

from autonomous import log
from autonomous.errors import DanglingReferenceError


class AutoEncoder:
    @classmethod
    def encode(cls, objs):
        try:
            if isinstance(objs, dict):
                obj_copy = {k: cls.encode(v) for k, v in objs.items()}
            elif isinstance(objs, list):
                obj_copy = [cls.encode(v) for v in objs]
            else:
                obj_copy = cls().default(objs)
            return obj_copy
        except DanglingReferenceError as e:
            log(e)
            return None

    def default(self, o):
        from autonomous.model.automodel import AutoModel, DelayedModel

        if issubclass(type(o), (AutoModel, DelayedModel)):
            name = "AutoModel"
        else:
            name = type(o).__name__

        encoder_name = f"encode_{name}"

        try:
            encoder = getattr(self, encoder_name)
        except AttributeError:
            return o
        else:
            encoded = {"__extended_json_type__": name, "value": encoder(o)}

        return encoded

    def encode_datetime(self, o):
        return o.isoformat()

    def encode_AutoModel(self, o):
        if o.pk:
            return {
                "pk": o.pk,
                "_automodel": o.model_name(qualified=True),
            }
        else:
            log(
                o,
                "The above object was not been saved. You must save subobjects if you want them to persist.",
            )
            raise ValueError("Cannot encode unsaved AutoModel")


class AutoDecoder:
    @classmethod
    def decode(cls, objs):
        decoder = cls()
        if isinstance(objs, dict):
            if "__extended_json_type__" in objs:
                objs = decoder.default(objs)
            else:
                for k, v in objs.items():
                    objs[k] = cls.decode(v)
        elif isinstance(objs, list):
            for i, v in enumerate(objs):
                objs[i] = cls.decode(v)
        return objs

    def default(self, obj):
        try:
            name = obj["__extended_json_type__"]
            decoder_name = f"decode_{name}"
            decoder = getattr(self, decoder_name)
        except (KeyError, AttributeError, TypeError):
            return obj
        else:
            return decoder(obj)

    def decode_datetime(self, o):
        return datetime.fromisoformat(o["value"])

    def decode_AutoModel(self, o):
        log(o)
        obj = o["value"]
        try:
            from autonomous.model.automodel import DelayedModel

            if not obj["pk"]:
                raise KeyError
            return DelayedModel(obj["_automodel"], obj["pk"])
        except KeyError:
            log(
                "AutoModel",
                "The above object was not been saved. You must save subobjects if you want them to persist.",
            )
            raise ValueError("Cannot decode unsaved AutoModel")



"""
    _summary_

_extended_summary_

:return: _description_
:rtype: _type_
"""

import glob
import os
import subprocess
import urllib.parse
from datetime import datetime

import pymongo

from autonomous import log

from .table import Table


class Database:
    """
     _summary_

    _extended_summary_

    :return: _description_
    :rtype: _type_
    """

    def __init__(
        self,
        host=os.getenv("DB_HOST", "db"),
        port=os.getenv("DB_PORT", 27017),
        password=os.getenv("DB_PASSWORD"),
        username=os.getenv("DB_USERNAME"),
        db=os.getenv("DB_DB"),
    ):
        """
        create an interface for your database
        """
        # log(self.username, self.password)
        username = urllib.parse.quote_plus(str(username))
        password = urllib.parse.quote_plus(str(password))
        self.connect_str = f"mongodb://{username}:{password}@{host}:{port}"
        # log(f"mongodb://{username}:{password}@{host}", port=int(port))
        self.db = pymongo.MongoClient(
            f"mongodb://{username}:{password}@{host}", port=int(port)
        )[db]
        self.tables = {}

    def get_table(self, table="default", schema=None):
        """
        opens the table from the file, which clears any changed data
        """
        if not self.tables.get(table):
            self.tables[table] = Table(table, schema, self.db)
        return self.tables[table]

    def dbdump(self, directory):
        """
        dumps the database to a json file
        """
        datetime_string = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        command_str = f'mongodump --uri="{self.connect_str}" --archive="{directory}/dbbackup-{datetime_string}.archive"'
        result = subprocess.Popen(command_str, shell=True).wait()
        log(result, command_str)

    def dbload(self, directory):
        """
        loads the database from a json file
        """
        files = glob.glob(
            f"{directory}/dbbackup-*.archive"
        )  # replace with your directory path

        # Find the file with the most recent timestamp
        latest_file = max(files, key=os.path.getctime)
        log(latest_file)
        command_str = (
            f'mongorestore --uri="{self.connect_str}" --archive="{latest_file}"'
        )
        result = subprocess.Popen(command_str, shell=True).wait()
        log(result, command_str)



# from .agents.autogen import AutoGenAgent
# from .agents.local import LocalAIAgent
import os

from .agents.mockai import MockAIAgent
from .agents.openai import OpenAIAgent


class AutoTeam:
    def __init__(self, model=None):
        if model:
            self.proxy = model()
        else:
            model = os.getenv("AI_AGENT", "openai")
            self.proxy = None
            if model == "openai":
                self.proxy = OpenAIAgent()
            elif model == "mockai":
                self.proxy = MockAIAgent()
            else:
                raise Exception("Invalid model")

    def generate_audio(self, prompt, **kwargs):
        return self.proxy.generate_audio(prompt, **kwargs)

    def generate_image(self, prompt, **kwargs):
        return self.proxy.generate_image(prompt, **kwargs)

    def generate_json(
        self,
        text,
        functions,
        name="json_agent",
        primer_text="",
        file_data=None,
        context=[],
    ):
        return self.proxy.generate_json(
            text,
            functions,
            name=name,
            primer_text=primer_text,
            file_data=file_data,
            context=context,
        )

    def generate_text(
        self, text, name="json_agent", primer_text="", file=None, context=[]
    ):
        return self.proxy.generate_text(
            text, name=name, primer_text=primer_text, file_data=file, context=context
        )

    def summarize_text(self, text, primer=""):
        return self.proxy.summarize_text(text, primer)



"""
    This module provides a User class that uses the OpenIDAuth class for authentication.
"""
from datetime import datetime

from autonomous import log
from autonomous.model.autoattribute import AutoAttribute
from autonomous.model.automodel import AutoModel


class AutoUser(AutoModel):
    """
    This class represents a user who can authenticate using OpenID.
    """

    attributes = {
        "name": AutoAttribute("TEXT"),
        "email": AutoAttribute("TEXT", required=True),
        "last_login": datetime.now(),
        "state": "unauthenticated",
        "provider": None,
        "role": "user",
        "token": None,
    }

    def __eq__(self, other):
        return self.pk == other.pk

    @classmethod
    def authenticate(cls, user_info, token=None):
        """
        Initiates the authentication process.
        Returns a redirect URL which should be used to redirect the user to the OpenID provider for authentication.
        """
        email = user_info["email"].strip()
        name = user_info["name"].strip()
        user = cls.find(email=email)
        if not user:
            log(f"Creating new user for {email}")
            user = cls(name=name, email=email)

        # parse user_info into a user object
        user.name = name
        user.email = email
        user.last_login = datetime.now()
        user.state = "authenticated"
        user.save()
        return user

    @property
    def is_authenticated(self):
        """
        Returns True if the user is authenticated, False otherwise.
        """
        return self.state == "authenticated"

    @classmethod
    def get_guest(cls):
        """
        Returns a guest user.
        """
        return cls(name="Guest", email="guest@mail.com", state="guest", role="guest")

    @property
    def is_guest(self):
        """
        Returns True if the user is a guest, False otherwise.
        """
        return self.state == "guest"



# opt : Optional[str]  # for optional attributes
# default : Optional[str] = "value"  # for default values
import copy
import importlib
import json
from abc import ABC
from datetime import datetime

from autonomous import log
from autonomous.errors import DanglingReferenceError

from .autoattribute import AutoAttribute
from .orm import ORM
from .serializer import AutoDecoder, AutoEncoder


class DelayedModel:
    def __init__(self, model, pk):
        # log(model, pk)
        assert model
        assert pk
        module_name, class_name = model.rsplit(".", 1)
        module = importlib.import_module(module_name)
        model = getattr(module, class_name)
        object.__setattr__(self, "_delayed_model", model)
        object.__setattr__(self, "_delayed_pk", pk)
        object.__setattr__(self, "_delayed_obj", None)

    def _instance(self):
        #### DO NOT TO ANY LOGGING IN THIS METHOD; IT CAUSES INFINITE RECURSION ####
        if not object.__getattribute__(self, "_delayed_obj"):
            _pk = object.__getattribute__(self, "_delayed_pk")
            _model = object.__getattribute__(self, "_delayed_model")
            _obj = _model.get(_pk)
            if not _pk or not _model or _obj is None:
                raise DanglingReferenceError(_model, _pk, _obj)
            else:
                object.__setattr__(self, "_delayed_obj", _obj)

        return object.__getattribute__(self, "_delayed_obj")

    # def __getattr__(self, name):
    #     return getattr(self._instance(), name)

    def __getattribute__(self, name):
        # log(name)
        if name in [
            "_delayed_model",
            "_delayed_pk",
            "_delayed_obj",
            "_instance",
        ]:
            return object.__getattribute__(self, name)
        try:
            return object.__getattribute__(self._instance(), name)
        except DanglingReferenceError as e:
            log(e)
            return None

    def __setattr__(self, name, value):
        if name.startswith("_delayed"):
            object.__setattr__(self, name, value)
        else:
            setattr(self._instance(), name, value)

    def __delattr__(self, name):
        delattr(self._instance(), name)

    def __nonzero__(self):
        return bool(self._instance())

    def __str__(self):
        return str(self._instance().__dict__)

    def __repr__(self):
        msg = f"\n<<DelayedModel {self._delayed_model.__name__}:{self._delayed_pk}>>"
        return msg

    def __hash__(self):
        return hash(self._instance())


class AutoModel(ABC):
    attributes = {}
    _table_name = ""
    _table = None
    _orm = ORM

    def __new__(cls, *args, **kwargs):
        """
        Create a new instance of the AutoModel.

        This method is responsible for creating a new instance of the AutoModel class.
        It sets default attributes, populates the object from the database if a primary key is provided,
        and handles additional keyword arguments.

        Args:
            cls: The class itself.
            *args: Positional arguments.
            **kwargs: Keyword arguments, including 'pk' for primary key.

        Returns:
            obj: The created AutoModel instance.
        """
        obj = super().__new__(cls)
        pk = kwargs.pop("pk", None)
        # set default attributes
        # Get model data from database
        result = cls.table().get(pk) or {}

        # set object attributes
        for k, v in cls.attributes.items():
            if isinstance(v, AutoAttribute):
                v = v.default
            setattr(obj, k, result.get(k, copy.deepcopy(v)))
        obj.pk = pk
        obj.__dict__ |= kwargs
        # breakpoint()
        obj.__dict__ = AutoDecoder.decode(obj.__dict__)
        obj._automodel = obj.model_name(qualified=True)
        obj.last_updated = datetime.now()

        return obj

    def __getattribute__(self, name):
        obj = super().__getattribute__(name)
        if not name.startswith("__"):
            if isinstance(obj, DelayedModel):
                try:
                    result = obj._instance()
                except DanglingReferenceError as e:
                    log(e)
                    super().__setattr__(name, None)
                    self.save()
                    return None
                else:
                    super().__setattr__(name, result)

            elif isinstance(obj, list):
                for i, item in enumerate(obj):
                    if isinstance(item, DelayedModel):
                        try:
                            result = item._instance()
                        except DanglingReferenceError as e:
                            log(e)
                            obj[i] = None
                            self.save()
                        else:
                            obj[i] = result

            elif isinstance(obj, dict):
                for key, item in obj.items():
                    if isinstance(item, DelayedModel):
                        try:
                            result = item._instance()
                        except DanglingReferenceError as e:
                            log(e)
                            obj[key] = None
                            self.save()
                        else:
                            obj[key] = result
        return obj

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Return a string representation of the AutoModel instance.

        Returns:
            str: A string representation of the AutoModel instance.
        """
        return str(self.__dict__)

    def __eq__(self, other):
        return self.pk == other.pk

    @classmethod
    def load_model(cls, model, module=None):
        if not module:
            module = f"models.{model.lower()}"
        module = importlib.import_module(module)
        return getattr(module, model)

    @classmethod
    def table(cls):
        # breakpoint()
        if not cls._table or cls._table.name != cls.__name__:
            cls.attributes["pk"] = None
            cls.attributes["last_updated"] = datetime.now()
            cls.attributes["_automodel"] = AutoAttribute(
                "TEXT", default=cls.model_name(qualified=True)
            )
            cls._table = cls._orm(cls._table_name or cls.__name__, cls.attributes)
        return cls._table

    @classmethod
    def model_name(cls, qualified=False):
        """
        Get the fully qualified name of this model.

        Returns:
            str: The fully qualified name of this model.
        """
        return f"{cls.__module__}.{cls.__name__}" if qualified else cls.__name__

    @property
    def _id(self):
        """
        Get the primary key of this model.

        Returns:
            int: The primary key of this model.
        """
        return self.pk

    @_id.setter
    def _id(self, _id):
        """
        Get the primary key of this model.

        Returns:
            int: The primary key of this model.
        """
        self.pk = str(_id)

    def validate(self):
        """
        Validate this model.
        """
        for key, vattr in self.attributes.items():
            if isinstance(vattr, AutoAttribute):
                val = getattr(self, key)
                if vattr.type == "TEXT":
                    if not isinstance(val, str):
                        raise TypeError(f"Value must be a string, not {type(val)}")
                elif vattr.type == "NUMERIC":
                    if not isinstance(val, (int, float)):
                        raise TypeError(f"Value must be a number, not {type(val)}")
                elif vattr.type == "MODEL":
                    # log(isinstance(val, (AutoModel, DelayedModel)), type(val))
                    if val is not None and not isinstance(
                        val, (AutoModel, DelayedModel)
                    ):
                        raise TypeError(
                            f"Value must be an AutoModel or None, not {type(val)}"
                        )
                else:
                    raise ValueError(f"Invalid type {self.type}")

                if vattr.required and val is None:
                    raise ValueError(f"{key} is required")
                if vattr.unique and len(self.search(**{key: val})) > 1:
                    raise ValueError(f"{key} must be unique")
                if vattr.primary_key:
                    self.pk = val

    def save(self):
        """
        Save this model to the database.

        Returns:
            int: The primary key (pk) of the saved model.
        """
        self.validate()
        serialized_obj = self.serialize()
        serialized_obj["pk"] = self.pk
        self.pk = self.table().save(serialized_obj)

        return self.pk

    @classmethod
    def get(cls, pk):
        """
        Get a model by primary key.

        Args:
            pk (int): The primary key of the model to retrieve.

        Returns:
            AutoModel or None: The retrieved AutoModel instance, or None if not found.
        """
        table = cls.table()
        result = table.get(pk)
        return cls(**result) if result else None

    @classmethod
    def random(cls):
        """
        Get a model by primary key.

        Args:
            pk (int): The primary key of the model to retrieve.

        Returns:
            AutoModel or None: The retrieved AutoModel instance, or None if not found.
        """
        result = cls.table().random()
        # breakpoint()
        return cls(**result) if result else None

    @classmethod
    def all(cls):
        """
        Get all models of this type.

        Returns:
            list: A list of AutoModel instances.
        """
        return [cls(**o) for o in cls.table().all()]

    @classmethod
    def search(cls, **kwargs):
        """
        Search for models containing the keyword values.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            list: A list of AutoModel instances that match the search criteria.
        """
        for k, v in kwargs.items():
            kwargs[k] = AutoEncoder.encode(v)
        return [cls(**attribs) for attribs in cls.table().search(**kwargs)]

    @classmethod
    def find(cls, **kwargs):
        """
        Find the first model containing the keyword values and return it.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            AutoModel or None: The first matching AutoModel instance, or None if not found.
        """
        attribs = cls.table().find(**kwargs)
        return cls(**attribs) if attribs else None

    def delete(self):
        """
        Delete this model from the database.
        """
        return self.table().delete(self.pk)

    def serialize(self):
        """
        Serialize this model to a dictionary.

        Returns:
            dict: A dictionary representation of the serialized model.
        """
        vars = {k: v for k, v in self.__dict__.items() if k in self.attributes}
        json_vars = AutoEncoder.encode(vars)
        return json_vars



import json
import os
import random
import time
from base64 import b64decode

from openai import NotFoundError, OpenAI

from autonomous import AutoModel, log


class OAIAgent(AutoModel):
    client = None
    attributes = {
        "model": "gpt-4-turbo-preview",
        "_agent_id": None,
        "messages": [],
        "tools": {},
        "name": "agent",
        "instructions": "You are highly skilled AI trained to assist with various tasks.",
        "description": "A helpful AI assistant trained to assist with various tasks.",
    }

    def __init__(self, **kwargs):
        self.client = OpenAI(api_key=os.environ.get("OPENAI_KEY"))

    @property
    def agent_id(self):
        if not self._agent_id:
            agent = self.client.beta.assistants.create(
                instructions=self.instructions,
                description=self.description,
                name=self.name,
                model=self.model,
            )
            self._agent_id = agent.id
            log(f"==== Creating Agent with ID: {self._agent_id} ====")
            self.save()
        return self._agent_id

    def clear_files(self, file_id=None):
        assistant_files = self.client.beta.assistants.files.list(
            assistant_id=self.agent_id
        )
        log(f"==== Files: {assistant_files.data}")
        if file_ids := [o.id for o in assistant_files.data]:
            if file_id and file_id in file_ids:
                file_ids = [file_id]
            for file_id in file_ids:
                self.client.beta.assistants.files.delete(
                    assistant_id=self.agent_id, file_id=file_id
                )
                self.client.files.delete(file_id=file_id)
            self.tools.pop("retrieval", None)
            self.save()
        return self.client.files.list()

    def attach_file(self, file_contents):
        self.tools["retrieval"] = {"type": "retrieval"}
        self.client.beta.assistants.update(
            self.agent_id,
            tools=list(self.tools.values()),
        )
        file_obj = self.client.files.create(file=file_contents, purpose="assistants")
        self.client.beta.assistants.files.create(
            assistant_id=self.agent_id, file_id=file_obj.id
        )
        self.save()
        return file_obj.id

    def _add_function(self, user_function):
        self.tools["function"] = {"type": "function", "function": user_function}
        self.client.beta.assistants.update(
            self.agent_id,
            tools=list(self.tools.values()),
        )
        return """
        IMPORTANT: always use the function 'response' tool to respond to the user with the requested JSON schema. Never add any other text to the response.
        """

    def _format_messages(self, messages):
        message_list = []
        if isinstance(messages, str):
            message_list.insert(0, {"role": "user", "content": messages})
        else:
            for message in messages:
                if isinstance(message, str):
                    message_list.insert(0, {"role": "user", "content": message})
                else:
                    raise Exception(
                        f"==== Invalid message: {message} ====\nMust be a string "
                    )
        return message_list

    def generate(self, messages, function=None):
        _instructions_addition = self._add_function(function) if function else ""

        formatted_messages = self._format_messages(messages)
        thread = self.client.beta.threads.create(messages=formatted_messages)

        run = self.client.beta.threads.runs.create(
            thread_id=thread.id,
            assistant_id=self.agent_id,
            additional_instructions=_instructions_addition,
        )

        while run.status in ["queued", "in_progress"]:
            run = self.client.beta.threads.runs.retrieve(
                thread_id=thread.id,
                run_id=run.id,
            )
            time.sleep(0.5)
            log(f"==== Job Status: {run.status} ====")

        if run.status in ["failed", "expired", "canceled"]:
            log(f"==== Error: {run.last_error} ====")
            return None

        result = None
        if run.status == "completed":
            response = self.client.beta.threads.messages.list(thread_id=thread.id)
            result = response.data[0].content[0].text.value
        elif run.status == "requires_action":
            results = run.required_action.submit_tool_outputs.tool_calls[
                0
            ].function.arguments
            result = results[results.find("{") : results.rfind("}") + 1]
            log(f"====result: {result} ====")
            try:
                result = json.loads(result, strict=False)
            except Exception:
                if result:
                    result = (
                        f"Rewrite the following response as valid JSON: \n\n {result}"
                    )
                    response = self.client.chat.completions.create(
                        model="gpt-4",
                        messages=[
                            {
                                "role": "system",
                                "content": "You are a helpful AI assistant trained to take unstructured data strings and make them valid structured JSON data. While you try to minimized data loss, valid JSON is more important than data integrity.",
                            },
                            {"role": "user", "content": result},
                        ],
                    )
                    result = json.loads(result, strict=False)
        else:
            log(f"====Status: {run.status} Error: {run.last_error} ====")
        return result

    def generate_audio(self, prompt, file_path, **kwargs):
        voice = kwargs.get("voice") or random.choice(
            ["alloy", "echo", "fable", "onyx", "nova", "shimmer"]
        )
        response = self.client.audio.speech.create(
            model="tts-1",
            voice=voice,
            input=prompt,
        )

        return response.stream_to_file(file_path)

    def generate_image(self, prompt, **kwargs):
        image = None
        try:
            response = self.client.images.generate(
                model="dall-e-3", prompt=prompt, response_format="b64_json", **kwargs
            )
            image_dict = response.data[0]
        except Exception as e:
            print(f"==== Error: Unable to create image ====\n\n{e}")
        else:
            image = b64decode(image_dict.b64_json)
        return image

    def summarize_text(self, text, primer=""):
        message = [
            {
                "role": "system",
                "content": f"You are a highly skilled AI trained in language comprehension and summarization.{primer}",
            },
            {"role": "user", "content": text},
        ]
        response = self.client.chat.completions.create(model="gpt-4", messages=message)
        try:
            result = response.choices[0].message.content
        except Exception as e:
            log(f"{type(e)}:{e}\n\n Unable to generate content ====")
            return None

        return result



"""
    _summary_

_extended_summary_

:return: _description_
:rtype: _type_
"""
import json
import random

from bson.objectid import ObjectId

from autonomous import log
from autonomous.model.autoattribute import AutoAttribute


class Table:
    def __init__(self, name, attributes, db):
        self._db = db[name]
        self.name = name
        # log(attributes)
        self._rules = {}

        for k, v in attributes.items():
            if isinstance(v, AutoAttribute):
                self._rules[k] = v
            elif isinstance(v, str):
                self._rules[k] = AutoAttribute("TEXT", default=v)
            elif isinstance(v, (int, float)):
                self._rules[k] = AutoAttribute("NUMERIC", default=v)
            else:
                self._rules[k] = None
        self._index = self._get_index(f"idx:{name}")

    def __str__(self):
        return json.dumps(self.all(), indent=4)

    def _get_index(self, name):
        pass

    def _validate(self, k, v):
        # log(k, v, self._rules)
        if rule := self._rules.get(k):
            if rule.type == "NUMERIC":
                if v:
                    try:
                        float(v)
                    except TypeError:
                        raise Exception(
                            f"VALIDATION ERROR: Invalid attribute value. Must be a number: {k}:{v}"
                        )

            if rule.required:
                try:
                    assert v is not None
                except AssertionError:
                    raise Exception(
                        f"VALIDATION ERROR: Attribute Required. Must not be 'None': {k}:{v}"
                    )

    def save(self, obj):
        for k, v in obj.items():
            try:
                self._validate(k, v)
            except Exception as e:
                # log(e)
                raise e

        if obj_id := obj.get("_id"):
            obj["_id"] = ObjectId(obj_id)
            self._db.replace_one({"_id": obj["_id"]}, obj, True)
        else:
            obj.pop("_id", None)
            obj["_id"] = self._db.insert_one(obj).inserted_id
        return str(obj["_id"])

    def count(self):
        return self._db.count_documents({})

    def delete(self, _id):
        try:
            return self._db.delete_one({"_id": ObjectId(_id)}).acknowledged
        except Exception as e:
            log(e)

    def _convert_to_dot_notation(self, search_terms, fuzzy_search=False, prefix=""):
        dot_notation = {}
        for key, value in search_terms.items():
            if isinstance(value, dict):
                dot_notation.update(
                    self._convert_to_dot_notation(
                        value, fuzzy_search=fuzzy_search, prefix=f"{prefix}{key}."
                    )
                )
            else:
                if fuzzy_search and isinstance(value, str):
                    dot_notation[f"{prefix}{key}"] = {"$regex": value, "$options": "i"}
                else:
                    dot_notation[f"{prefix}{key}"] = value
        return dot_notation

    def find(self, **search_terms):
        search_terms = self._convert_to_dot_notation(search_terms)
        result = self._db.find_one(search_terms)
        if result:
            result["_id"] = str(result["_id"])
            return result

    def search(self, **search_terms):
        fuzzy_search = search_terms.pop("_FUZZY_SEARCH", False)
        search_terms = self._convert_to_dot_notation(
            search_terms, fuzzy_search=fuzzy_search
        )
        result = self._db.find(search_terms) or []

        objs = []
        for o in result:
            o["_id"] = str(o["_id"])
            objs.append(o)
        # log(search_terms, fuzzy_search, objs)
        return objs

    def get(self, _id):
        if not _id or _id == "None":
            return None
        if o := self._db.find_one({"_id": ObjectId(_id)}):
            o["_id"] = str(o["_id"])
            return o

    def all(self):
        objs = []
        for o in self._db.find():
            o["_id"] = str(o["_id"])
            objs.append(o)
        return objs

    def random(self):
        keys = [o for o in self._db.find({}, projection=["_id"])]
        # log(keys)
        try:
            key = random.choice(keys)
        except Exception as e:
            # log(e, f"Table '{self.name}' is empty.")
            return None
        else:
            result = self.get(str(key["_id"]))
            # log(result)
            return result

    def clear(self):
        # breakpoint()
        return self._db.drop()
